from .completion import autocomplete_only_nodes
from .locals import decimal_separator, thousands_separator
from .queries import (
    clever_parsing,
    include_parents,
    query_exceptions,
    recursive_queries,
)



import json
from jsonutils.exceptions import JSONQueryException, JSONQueryMultipleValues
import unittest
from datetime import datetime

import pytz
from jsonutils.base import (
    JSONBool,
    JSONCompose,
    JSONDict,
    JSONFloat,
    JSONInt,
    JSONList,
    JSONNode,
    JSONNull,
    JSONObject,
    JSONPath,
    JSONSingleton,
    JSONStr,
)
from jsonutils.encoders import JSONObjectEncoder
from jsonutils.query import All, QuerySet, SingleQuery


class JsonTest(unittest.TestCase):
    def setUp(self):
        self.test1 = JSONObject(
            [
                {"Float": 2.3, "Int": 1, "Str": "string"},
                {"Dict": {"Float": 0.0, "List": [1, 2, 3]}},
            ]
        )
        self.test2 = JSONObject(True)
        self.test3 = JSONObject(
            {"List": (True, False), "Bool": True, "Dict": {"Float": 3.2}}
        )
        self.test4 = JSONObject(
            {"List": [0, 0.1, "str", None], "Dict": {"Bool": True, "None": None}}
        )

        self.test5 = JSONObject(
            [
                {
                    "Float": 1.1,
                    "Dict": {
                        "List": [{"Str": "string1", "List": [None, True, False, 1]}],
                        "Null": None,
                    },
                },
                {
                    "List": [{"Dict": {"Float": "1.2"}, "Bool": True}],
                    "Dict": {"List": [None, {"Str": "string2"}]},
                },
                {"Datetime": "2021-05-01", "Dict": {"Datetime": "2021/06/01"}},
            ]
        )

        self.test6 = JSONObject(
            {
                "position_data": [
                    {"text": "dummy text 1", "pos": [1, 2]},
                    {"text": "dummy text 2", "pos": [3, 2]},
                    {"text": "dummy text 3", "pos": [1, 4]},
                    {"text": "dummy text 4", "pos": [2, 5]},
                    {"text": "dummy text 5", "pos": [4, 1]},
                    {"text": "dummy text 6", "pos": [1, 1, 5]},
                ],
                "timestamp_data": [
                    {"value": 523687, "timestamp": "2021-05-01 08:00:00"},
                    {"value": 523689, "timestamp": "2021-05-01 09:00:00"},
                    {"value": 523787, "timestamp": "2021-05-02 08:30:00"},
                    {"value": 525687, "timestamp": "2021-05-05 18:00:25"},
                ],
                "boolean_data": [
                    {"data": [True, 1]},
                    {"data": [False, None]},
                    {"data": [0, 1]},
                    {"data": (False, True)},
                ],
            }
        )

        self.test7 = JSONObject(
            {
                "candidatos": [
                    {
                        "Nombre completo": "Fhànçaömá",
                        "Pasión": "Música",
                        "€uros": 356.3,
                    },
                    {"Nombre completo": "Üm€", "Pasión": "Comida", "€€": "28.36K €"},
                ]
            }
        )

        self.test8 = JSONObject(
            {
                "data_list": [
                    {"name": "", "id": 0, "list": [], "tuple": (), "none": None},
                    {},
                ],
                "list": [0, 1, 2],
            }
        )

    def test_unknown_types(self):
        test = JSONObject({"data": {"datetime": datetime(2021, 1, 1, 0, 0, 0)}})

        self.assertEqual(test.data.datetime, JSONStr("2021-01-01T00:00:00"))

    def test_types(self):
        """Assert all child types are the correct ones"""

        self.assertIsInstance(self.test1, JSONList)
        self.assertIsInstance(self.test1[0], JSONDict)
        self.assertIsInstance(self.test1[0]["Float"], JSONFloat)
        self.assertIsInstance(self.test1[0]["Int"], JSONInt)
        self.assertIsInstance(self.test1[0]["Str"], JSONStr)
        self.assertIsInstance(self.test1[1], JSONDict)
        self.assertIsInstance(self.test1[1]["Dict"], JSONDict)
        self.assertIsInstance(self.test1[1]["Dict"]["Float"], JSONFloat)
        self.assertIsInstance(self.test1[1]["Dict"]["List"], JSONList)

        self.assertIsInstance(self.test3, JSONDict)
        self.assertIsInstance(self.test3["List"], JSONList)
        self.assertIsInstance(self.test3["List"][0], JSONBool)
        self.assertIsInstance(self.test3["List"][1], JSONBool)
        self.assertIsInstance(self.test3["Bool"], JSONBool)
        self.assertIsInstance(self.test3["Dict"], JSONDict)
        self.assertIsInstance(self.test3["Dict"]["Float"], JSONFloat)

        self.assertIsInstance(self.test4, JSONDict)
        self.assertIsInstance(self.test4["List"], JSONList)
        self.assertIsInstance(self.test4["List"][0], JSONInt)
        self.assertIsInstance(self.test4["List"][1], JSONFloat)
        self.assertIsInstance(self.test4["List"][2], JSONStr)
        self.assertIsInstance(self.test4["List"][3], JSONNull)
        self.assertIsInstance(self.test4["Dict"], JSONDict)
        self.assertIsInstance(self.test4["Dict"]["Bool"], JSONBool)
        self.assertIsInstance(self.test4["Dict"]["None"], JSONNull)

    def test_append_method(self):

        test1 = self.test1.copy()

        test1.append({"fake": True})

        self.assertIsInstance(test1, JSONList)
        self.assertIsInstance(test1[0], JSONDict)
        self.assertIsInstance(test1[0]["Float"], JSONFloat)
        self.assertIsInstance(test1[0]["Int"], JSONInt)
        self.assertIsInstance(test1[0]["Str"], JSONStr)
        self.assertIsInstance(test1[1], JSONDict)
        self.assertIsInstance(test1[1]["Dict"], JSONDict)
        self.assertIsInstance(test1[1]["Dict"]["Float"], JSONFloat)
        self.assertIsInstance(test1[1]["Dict"]["List"], JSONList)
        self.assertIsInstance(test1[2], JSONDict)
        self.assertIsInstance(test1[2]["fake"], JSONBool)

        self.assertEqual(test1[0].parent, test1)
        self.assertEqual(test1[1].parent, test1)
        self.assertEqual(test1[2].parent, test1)

        self.assertEqual(test1[0]._index, 0)
        self.assertEqual(test1[1]._index, 1)
        self.assertEqual(test1[2]._index, 2)

        self.assertEqual(test1[0]["Float"]._key, "Float")
        self.assertEqual(test1[2]["fake"]._key, "fake")
        self.assertEqual(test1[2]["fake"].jsonpath, JSONPath("2/fake/"))

        self.assertTrue(test1.query(fake=True).exists())

    def test_list_set_items(self):

        test1 = self.test1.copy()

        test1[1] = ["A", {"B": None}]

        self.assertIsInstance(test1, JSONList)
        self.assertIsInstance(test1[0], JSONDict)
        self.assertIsInstance(test1[0]["Float"], JSONFloat)
        self.assertIsInstance(test1[0]["Int"], JSONInt)
        self.assertIsInstance(test1[0]["Str"], JSONStr)
        self.assertIsInstance(test1[1], JSONList)
        self.assertIsInstance(test1[1][0], JSONStr)
        self.assertIsInstance(test1[1][1], JSONDict)
        self.assertIsInstance(test1[1][1]["B"], JSONNull)

    def test_dict_set_items(self):

        test3 = self.test3.copy()

        test3["new_key"] = {"key": [None, "None"]}

        self.assertIsInstance(test3, JSONDict)
        self.assertIsInstance(test3["List"], JSONList)
        self.assertIsInstance(test3["List"][0], JSONBool)
        self.assertIsInstance(test3["List"][1], JSONBool)
        self.assertIsInstance(test3["Bool"], JSONBool)
        self.assertIsInstance(test3["Dict"], JSONDict)
        self.assertIsInstance(test3["Dict"]["Float"], JSONFloat)
        self.assertIsInstance(test3["new_key"], JSONDict)
        self.assertIsInstance(test3["new_key"]["key"], JSONList)
        self.assertIsInstance(test3["new_key"]["key"][0], JSONNull)
        self.assertIsInstance(test3["new_key"]["key"][1], JSONStr)

        self.assertEqual(test3["List"].parent, test3)
        self.assertEqual(test3["Bool"].parent, test3)
        self.assertEqual(test3["new_key"].parent, test3)
        self.assertEqual(test3["new_key"]["key"].parent, test3["new_key"])
        self.assertEqual(test3["new_key"]["key"][0].parent, test3["new_key"]["key"])

        self.assertEqual(
            test3["new_key"]["key"][0].jsonpath, JSONPath("new_key/key/0/")
        )

    def test_keys(self):

        self.assertEqual(self.test1[0]._key, None)
        self.assertEqual(self.test1[1]._key, None)
        self.assertEqual(self.test1[0]["Float"]._key, "Float")
        self.assertEqual(self.test1[0]["Int"]._key, "Int")
        self.assertEqual(self.test1[0]["Str"]._key, "Str")
        self.assertEqual(self.test1[1]["Dict"]._key, "Dict")
        self.assertEqual(self.test1[1]["Dict"]["Float"]._key, "Float")
        self.assertEqual(self.test1[1]["Dict"]["List"]._key, "List")

    def test_parents(self):
        """Check every child object has the right parent object"""

        self.assertEqual(self.test1.parent, None)
        self.assertEqual(
            self.test1[0].parent,
            JSONList(
                [
                    {"Float": 2.3, "Int": 1, "Str": "string"},
                    {"Dict": {"Float": 0.0, "List": [1, 2, 3]}},
                ]
            ),
        )
        self.assertEqual(
            self.test1[1].parent,
            JSONList(
                [
                    {"Float": 2.3, "Int": 1, "Str": "string"},
                    {"Dict": {"Float": 0.0, "List": [1, 2, 3]}},
                ]
            ),
        )
        self.assertEqual(
            self.test1[0]["Float"].parent,
            JSONDict({"Float": 2.3, "Int": 1, "Str": "string"}),
        )
        self.assertEqual(
            self.test1[1]["Dict"].parent,
            JSONDict({"Dict": {"Float": 0.0, "List": [1, 2, 3]}}),
        )
        self.assertEqual(
            self.test1[1]["Dict"]["Float"].parent,
            JSONDict({"Float": 0.0, "List": [1, 2, 3]}),
        )
        self.assertEqual(
            self.test1[1]["Dict"]["List"].parent,
            JSONDict({"Float": 0.0, "List": [1, 2, 3]}),
        )
        self.assertEqual(
            self.test1[1]["Dict"]["List"][0].parent,
            JSONList([1, 2, 3]),
        )
        self.assertEqual(
            self.test1[1]["Dict"]["List"][1].parent,
            JSONList([1, 2, 3]),
        )

    def test_roots_in_queryset(self):
        self.assertEqual(self.test6.query(data__1=True)._root, self.test6)
        self.assertEqual(
            self.test6.position_data.query(pos__1="44")._root, self.test6.position_data
        )

    def test_root(self):
        self.assertEqual(self.test6.position_data._0.root, self.test6)
        self.assertEqual(self.test6.position_data._1.root, self.test6)
        self.assertEqual(self.test6.root, self.test6)

    def test_paths(self):

        self.assertEqual(
            self.test1.query(List__contains=2).first().jsonpath,
            JSONPath("1/Dict/List/"),
        )
        self.assertEqual(
            self.test1.query(List__contains=2).first().jsonpath.expr,
            '[1]["Dict"]["List"]',
        )

    def test_json_serializable(self):
        """Assert that the JSONObject is serializable"""

        self.assertEqual(
            json.dumps(self.test1).replace('"', "'"), self.test1.__repr__()
        )
        self.assertEqual(
            json.dumps(self.test2, cls=JSONObjectEncoder).replace('"', "'"),
            self.test2.__repr__().lower(),
        )
        self.assertEqual(
            json.dumps(self.test3, cls=JSONObjectEncoder).replace('"', "'"),
            self.test3.__repr__().replace("True", "true").replace("False", "false"),
        )

        self.assertEqual(
            json.dumps(self.test4, cls=JSONObjectEncoder),
            '{"List": [0, 0.1, "str", null], "Dict": {"Bool": true, "None": null}}',
        )

    def test_get(self):
        test = self.test6
        test2 = self.test1

        self.assertEqual(
            test.get(timestamp="2021-05-01 08:00:00"),
            test.query(timestamp="2021-05-01 08:00:00").first(),
        )

        self.assertRaises(JSONQueryMultipleValues, lambda: test.get(pos__0=1))

        self.assertEqual(test2.get(Float=2.3), test2.query(Float="2.3").first())
        self.assertRaises(JSONQueryMultipleValues, lambda: test2.get(Float__gt="-1"))
        self.assertRaises(JSONQueryException, lambda: test2.get(Float__gt="-1"))

        self.assertRaisesRegex(
            JSONQueryException,
            "The query has not returned any result",
            lambda: test.get(timestamp="1895-01-01"),
        )

    def test_queries(self):

        self.assertEqual(self.test3.query(Bool="true"), [JSONBool(True)])
        self.assertEqual(self.test3.query(Bool__type=str), [])
        self.assertEqual(self.test3.query(Bool__type=bool), [JSONBool(True)])
        self.assertEqual(self.test4.query(None__type=None), [JSONNull(None)])
        self.assertEqual(self.test5.query(Str__length=7), ["string1", "string2"])
        self.assertEqual(self.test5.query(Str__length=0), [])
        self.assertRaises(JSONQueryException, lambda: self.test5.query(Str__length="1"))
        self.assertEqual(
            self.test3.query(List__contains=True), [[JSONBool(True), JSONBool(False)]]
        )
        self.assertEqual(
            self.test1.query(Dict__contains=["Float", "List"]).last(),
            {"Float": 0.0, "List": [1, 2, 3]},
        )
        self.assertEqual(
            self.test4.query(Dict__in=("Bool", "None", "Other")),
            [{"Bool": True, "None": None}],
        )
        self.assertEqual(
            self.test4.query(Dict__in=("Bool", "Other")),
            [],
        )
        self.assertEqual(
            self.test4.query(List__in=("0", "0.1", "str", None)),
            [[0, 0.1, "str", None]],
        )
        self.assertEqual(
            self.test5.query(Bool__contains=True),
            [True],
        )
        self.assertEqual(self.test5.query(Float=1.2), QuerySet([JSONStr(1.2)]))
        self.assertEqual(
            self.test5.query(Float__gt=1), QuerySet([JSONFloat(1.1), JSONStr(1.2)])
        )
        self.assertEqual(self.test5.query(Float__gt=2), QuerySet())
        self.assertEqual(self.test5.query(Str__exact="string2"), QuerySet(["string2"]))
        self.assertEqual(
            self.test5.query(List=[None, True, False, 1], include_parent_=True),
            [JSONDict({"Str": "string1", "List": [None, True, False, 1]})],
        )
        self.assertEqual(
            self.test5.query(Str__exact="string2", include_parent_=True),
            [{"Str": "string2"}],
        )
        self.assertEqual(
            self.test5.query(List__0__contains=("Dict", "Bool")),
            [[{"Dict": {"Float": "1.2"}, "Bool": True}]],
        )
        self.assertEqual(
            self.test5.query(List__0__contains="Str"),
            [[{"Str": "string1", "List": [None, JSONBool(True), False, 1]}]],
        )
        self.assertEqual(
            self.test5.query(Datetime__gt="2021-05-01"), QuerySet(["2021/06/01"])
        )
        self.assertEqual(
            self.test5.query(Datetime__contains="2021-05"),
            QuerySet([JSONStr("2021-05-01")]),
        )
        self.assertEqual(
            self.test5.query(List__parent__contains="Str"),
            [
                [JSONNull(None), JSONBool(True), JSONBool(False), 1],
            ],
        )
        self.assertEqual(
            self.test6.query(
                timestamp__gt="2021-05-01 08:30:00",
                timestamp__lte="2021-05-02 08:30:00",
                include_parent_=True,
            ),
            [
                {"value": 523689, "timestamp": "2021-05-01 09:00:00"},
                {"value": 523787, "timestamp": "2021-05-02 08:30:00"},
            ],
        )
        self.assertEqual(self.test6.query(pos__0__gte=2), [[3, 2], [2, 5], [4, 1]])
        self.assertEqual(
            self.test6.query(pos__2__gt=2, include_parent_=True),
            [{"text": "dummy text 6", "pos": [1, 1, 5]}],
        )
        self.assertEqual(self.test6.query(pos__5__gte=2), [])
        self.assertEqual(
            self.test6.query(pos__0__gte=2, pos__1__lt=5), [[3, 2], [4, 1]]
        )
        self.assertEqual(
            self.test6.query(pos__1__in=(1, 4), include_parent_=True),
            [
                {"text": "dummy text 3", "pos": [1, 4]},
                {"text": "dummy text 5", "pos": [4, 1]},
                {"text": "dummy text 6", "pos": [1, 1, 5]},
            ],
        )
        self.assertEqual(self.test8.query(list__isnull=True), QuerySet([[]]))
        self.assertEqual(self.test8.query(tuple__isnull=True), QuerySet([[]]))
        self.assertEqual(self.test8.query(none__isnull=True), QuerySet([None]))
        self.assertEqual(self.test8.query(id__isnull=True), QuerySet())
        self.assertEqual(self.test6.query(data__1=True), QuerySet([[False, True]]))

    def test_multiqueries(self):
        test = JSONObject(
            [
                {
                    "data": {
                        "data": [
                            1,
                            2,
                            {"data": 1, "date": "2021/08/01  12:00:00Z"},
                            {"data": {"data": 1, "date": "2021-08-02"}},
                        ],
                        "date": "2021/08/01  13:00:00.003Z",
                    },
                    "date": datetime(2021, 9, 8),
                }
            ]
        )

        self.assertEqual(
            self.test6.query(
                text__regex=r"(?:2|5)", pos__0__gte=2, include_parent_=True
            ),
            [
                {"text": "dummy text 2", "pos": [3, 2]},
                {"text": "dummy text 5", "pos": [4, 1]},
            ],
        )

        self.assertEqual(
            test.query(data=1, date__gte=datetime(2021, 8, 1), include_parent_=True),
            [
                {"data": 1, "date": "2021/08/01  12:00:00Z"},
                {"data": 1, "date": "2021-08-02"},
            ],
        )
        self.assertEqual(
            test.query(data__c_data__contains=1),
            [
                {
                    "data": [
                        1,
                        2,
                        {"data": 1, "date": "2021/08/01  12:00:00Z"},
                        {"data": {"data": 1, "date": "2021-08-02"}},
                    ],
                    "date": "2021/08/01  13:00:00.003Z",
                },
                {"data": 1, "date": "2021-08-02"},
            ],
        )
        self.assertEqual(
            test.annotate(date=None).query(
                data__c_data__contains=1, date__isnull=True, include_parent_=True
            ),
            [{"data": {"data": 1, "date": "2021-08-02"}, "date": None}],
        )

    def test_queries_traversing(self):
        # {
        #     "position_data": [
        #         {"text": "dummy text 1", "pos": [1, 2]},
        #         {"text": "dummy text 2", "pos": [3, 2]},
        #         {"text": "dummy text 3", "pos": [1, 4]},
        #         {"text": "dummy text 4", "pos": [2, 5]},
        #         {"text": "dummy text 5", "pos": [4, 1]},
        #         {"text": "dummy text 6", "pos": [1, 1, 5]},
        #     ],
        #     "timestamp_data": [
        #         {"value": 523687, "timestamp": "2021-05-01 08:00:00"},
        #         {"value": 523689, "timestamp": "2021-05-01 09:00:00"},
        #         {"value": 523787, "timestamp": "2021-05-02 08:30:00"},
        #         {"value": 525687, "timestamp": "2021-05-05 18:00:25"},
        #     ],
        # "boolean_data": [
        #     {"data": [True, 1]},
        #     {"data": [False, None]},
        #     {"data": [0, 1]},
        #     {"data": (False, True)},
        # ],
        # }
        test = self.test6

        self.assertListEqual(
            test.query(
                data__parent__parent__0__c_data__1__exact="1", include_parent_=True
            ),
            [
                {"data": [True, 1]},
                {"data": [False, None]},
                {"data": [0, 1]},
                {"data": [False, True]},
            ],
        )
        self.assertEqual(
            test.query(
                timestamp__lte="2021-05-01T23:59:00",
                timestamp__parent__c_value__contains=89,
                include_parent_=True,
            ),
            [{"value": 523689, "timestamp": "2021-05-01 09:00:00"}],
        )

    def test_single_query_exact(self):

        test_dict = JSONObject({"A": {"A": 1, "B": True}})
        test_list = JSONObject({"A": [1, "2", True, "false"]})
        test_str = JSONObject({"A": "lorep ipsum", "B": {"B1": [1, 2]}})
        test_bool = JSONObject({"A": True})

        self.assertTrue(
            SingleQuery("A", {"A": 1, "B": True})._check_against_node(test_dict.A)
        )
        self.assertFalse(SingleQuery("A", ["A", "B"])._check_against_node(test_dict.A))

        self.assertTrue(SingleQuery("A", "lorep ipsum")._check_against_node(test_str.A))
        self.assertFalse(SingleQuery("A", "lorepipsum")._check_against_node(test_str.A))
        self.assertTrue(
            SingleQuery("A__exact", "lorep ipsum")._check_against_node(test_str.A)
        )
        self.assertFalse(
            SingleQuery("A__exact", "lorepipsum")._check_against_node(test_str.A)
        )

        self.assertTrue(
            SingleQuery("A", [1, "2", True, "false"])._check_against_node(test_list.A)
        )
        self.assertTrue(
            SingleQuery("A", [1, 2, True, False])._check_against_node(test_list.A)
        )
        self.assertTrue(
            SingleQuery(
                "B1__parent__parent__c_A__exact", "lorep ipsum"
            )._check_against_node(test_str.B.B1)
        )

    def test_single_query_gt(self):
        test_float = JSONObject({"A": 1.3})
        test_list = JSONObject({"A": [2, 3], "B": []})

        self.assertTrue(SingleQuery("A__gt", 1)._check_against_node(test_float.A))
        self.assertTrue(SingleQuery("A__gt", [1, 2])._check_against_node(test_list.A))
        self.assertTrue(
            SingleQuery("A__gt", [1, 2, 5])._check_against_node(test_list.A)
        )
        self.assertTrue(SingleQuery("A__gt", [1])._check_against_node(test_list.A))
        self.assertFalse(SingleQuery("A__gt", 1)._check_against_node(test_list.A))
        self.assertFalse(SingleQuery("A__gt", [])._check_against_node(test_list.A))
        self.assertFalse(SingleQuery("A__gt", type)._check_against_node(test_list.A))
        self.assertFalse(SingleQuery("A__gt", [3, 2])._check_against_node(test_list.A))
        self.assertFalse(
            SingleQuery("A__gt", [1, 3, 5])._check_against_node(test_list.A)
        )
        self.assertFalse(
            SingleQuery("B__gt", [1, 3, 5])._check_against_node(test_list.B)
        )
        self.assertFalse(SingleQuery("B__gt", [])._check_against_node(test_list.B))

    def test_single_queries(self):
        test = JSONObject(
            [{"A": [1, 2], "B": {"A": "123"}}, {"date": "2021-05-04T09:08:00"}]
        )

        q1 = SingleQuery("A__contains", 1)
        q2 = SingleQuery("A__in", (0, 2, 1, 3))
        q3 = SingleQuery("A__in", (0, 2))
        q4 = SingleQuery("A__contains", 3)
        q5 = SingleQuery("date", datetime(2021, 5, 4, 9, 8, 0, tzinfo=pytz.utc))
        q6 = SingleQuery("B__c_A", 123)

        self.assertTrue(q1._check_against_node(test._0.A))
        self.assertTrue(q2._check_against_node(test._0.A))
        self.assertFalse(q3._check_against_node(test._0.A))
        self.assertTrue(q4._check_against_node(test._0.B.A))
        self.assertTrue(q5._check_against_node(test._1.date))
        self.assertTrue(q6._check_against_node(test._0.B))

    def test_update(self):

        test = JSONObject(
            {
                "data": {
                    "cik": "0005852352",
                    "fileNumber": "052-065820",
                    "team": [
                        {"name": "David", "bio": "An enthusiast", "date": "2021-05-08"},
                        {"name": "Alex", "bio": "A musician", "date": "2020-01-03"},
                    ],
                },
                "comments": [
                    {
                        "date": "2021-01-01T08:00:00",
                        "text": "Ouuu yeahhh",
                        "responses": 0,
                    },
                    {
                        "date": "2021-01-02T09:30:00",
                        "text": "That sounds quite good",
                        "responses": 22,
                    },
                ],
            }
        )
        test6 = self.test6.copy()

        test6.query(data__0=True).update("OK")
        test.data.team.query(name__contains="e").update("Veronica")

        self.assertEqual(test.data.team.query(name__contains="e"), ["Veronica"])
        self.assertEqual(test6.query(data__0=True), [])
        self.assertEqual(test6.query(data="OK"), ["OK"])

    def test_all_queries(self):
        test = JSONObject({"A": {"A": 1, "B": 2}, "B": ["A", {"A": 2}]})

        self.assertEqual(test.query(A=All), [{"A": 1, "B": 2}, "$1", 2])
        self.assertEqual(test.query(B=All), [2, ["A", {"A": 2}]])

    def test_setters(self):

        test = JSONObject({"data": [{"A": 1, "B": 2}]})
        test2 = JSONObject({"key": "mykey", "index": 1, "nested": {"index": "2"}})
        test3 = JSONObject([1, 2, {"A": [3, 4]}])

        self.assertEqual(
            test.query(data__0__contains="B"), QuerySet([[{"A": 1, "B": 2}]])
        )

        test.data._0.B = 555

        self.assertFalse(test.query(B=2).exists())
        self.assertEqual(test.query(B=555), [555])

        self.assertEqual(test2.key, "mykey")
        self.assertEqual(test2.index, "1")
        self.assertEqual(test2.nested.index, 2)

        test2.key = 111
        test2.index = 222
        test2.nested.index = 333

        self.assertDictEqual(
            test2, JSONObject({"key": 111, "index": 222, "nested": {"index": 333}})
        )
        self.assertEqual(test2.query(key=All), [111])
        self.assertListEqual(test2.query(index=All), [333, 222])
        self.assertEqual(test2.query(index=333).first().jsonpath, "nested/index/")

        self.assertTrue(test3.query(A__contains=4).exists())

        test3._0 = "first"
        test3._1 = "second"
        test3._2.A._1 = "last"

        self.assertListEqual(test3, JSONObject(["first", "second", {"A": [3, "last"]}]))
        self.assertFalse(test3.query(A__contains=4).exists())
        self.assertTrue(test3.query(A__contains="last").exists())

    def test_annotations(self):

        test = self.test1.copy()
        test2 = self.test1.copy()
        test3 = JSONObject([1, 2, {"A": [1, 2, {"B": 3}], "B": 4}])

        self.assertEqual(
            test.annotate(a1=1, a2=2),
            JSONObject(
                [
                    {"Float": 2.3, "Int": 1, "Str": "string", "a1": 1, "a2": 2},
                    {
                        "Dict": {"Float": 0.0, "List": [1, 2, 3], "a1": 1, "a2": 2},
                        "a1": 1,
                        "a2": 2,
                    },
                ]
            ),
        )

        self.assertEqual(test._1.Dict.a2.jsonpath, "1/Dict/a2/")
        self.assertEqual(test.query(a1=1).count(), 3)
        self.assertNotEqual(test, self.test1)

        self.assertEqual(
            test2.annotate(a1={"status": "OK"}).query(a1__contains="status"),
            [{"status": "OK"}, {"status": "OK"}, {"status": "OK"}],
        )

        self.assertEqual(
            test3.annotate(C=1, D=2),
            JSONObject(
                [1, 2, {"A": [1, 2, {"B": 3, "C": 1, "D": 2}], "B": 4, "C": 1, "D": 2}]
            ),
        )

        # now we remove the annotations and check if recovers original object

        test._remove_annotations()

        self.assertEqual(test, self.test1)
        self.assertFalse(test.query(a1=All).exists())

    def test_pop(self):

        test = JSONObject({"data": [{"name": "Dan", "age": 30}]})

        self.assertSetEqual(set(test.data._0._child_objects.values()), {"Dan", 30})
        self.assertEqual(test.data._0.name, "Dan")

        # remove a child

        test.data._0.pop("name")

        self.assertSetEqual(set(test.data._0._child_objects.values()), {30})
        self.assertFalse(test.query(name=All).exists())
        self.assertRaises(AttributeError, lambda: test.data._0.name)



import unittest

from jsonutils.utils.dict import TranslationDict


class JsonTest(unittest.TestCase):
    def test_default_keys(self):

        test = TranslationDict(a=1, b=2)
        self.assertEqual(test["a"], 1)
        self.assertEqual(test["b"], 2)
        self.assertEqual(test["c"], "c")
        self.assertEqual(test[1], 1)
        self.assertEqual(test[object], object)

    def test_default_custom(self):

        test = TranslationDict({"a": 1, "b": 2, "in": 3}, default_="OK")
        self.assertEqual(test["a"], 1)
        self.assertEqual(test["b"], 2)
        self.assertEqual(test["in"], 3)
        self.assertEqual(test["c"], "OK")
        self.assertEqual(test[1], "OK")
        self.assertEqual(test[object], "OK")



import re
from datetime import date, datetime
from typing import Union

from jsonutils.exceptions import JSONQueryException


class SingleQuery:
    """
    This class represents a single query object.
    It consists of a single key and a collection of actions that operate on that key.

    Arguments:
    ---------
        query_key: key part of the query. Must be an string
        query_value: target value of the query
    """

    def __init__(
        self,
        query_key: str,
        query_value: Union[float, int, str, None, bool, dict, list, tuple, datetime],
    ):

        if not isinstance(
            query_value,
            (
                type,
                float,
                int,
                str,
                type(None),
                bool,
                dict,
                list,
                tuple,
                date,
                datetime,
                AllChoices,
            ),
        ):
            raise JSONQueryException(
                f"Target value of query has invalid type: {type(query_value)}. Valid types are: float, int, str, None, bool, dict, list, tuple, date, datetime, allchoices"
            )

        self.target_value = query_value
        self._parse_key(query_key)

    def _parse_key(self, query_key):
        """Convert query full key in a list of actions"""

        splitted_query = [i for i in query_key.split("__") if i]

        if not splitted_query:
            raise JSONQueryException("Bad query. Missing target key")

        self.target_key = splitted_query[0]
        self.target_actions = splitted_query[1:] or ["exact"]

    def _check_against_node(self, node):
        """
        Check this single query against a target node object.
        node argument must be an instance of JSONNode.
        """
        from jsonutils.base import JSONNode

        if not isinstance(node, JSONNode):
            raise JSONQueryException(
                f"child argument must be JSONNode type, not {type(node)}"
            )
        # if child key does not match the target query key, returns False, because this is a single query
        if node._key != self.target_key:
            return False

        # grab a list of node actions, without the action suffix
        node_actions = [
            i.replace("_action", "") for i in dir(JSONNode) if i.endswith("action")
        ]
        obj = node

        actions_count = len(self.target_actions)
        for idx, action in enumerate(self.target_actions):

            # ---- MODIFICATORS ----
            if action == "parent":
                obj = obj.parent
                if obj is None:
                    return False
                if idx == actions_count - 1:
                    action = "exact"  # if parent is last action, take exact as the default one
                else:
                    continue  # continue to next action
            elif match := re.fullmatch(r"c_(\w+)", action):  # child modificator
                try:
                    obj = obj.__getitem__(match.group(1))
                except Exception:
                    return False
                if idx == actions_count - 1:
                    action = "exact"  # if child is last action, take exact as the default one
                else:
                    continue  # continue to next action
            elif action.isdigit():
                if not isinstance(obj, list):
                    return False
                try:
                    obj = obj[int(action)]
                except IndexError:
                    return False
                if idx == actions_count - 1:
                    action = "exact"  # if digit is last action, take exact as the default one
                else:
                    continue  # continue to next action

            # ---- ACTIONS ----
            # node actions can't interfer with modificators
            if action in node_actions:  # call corresponding node method
                return getattr(obj, action + "_action")(self.target_value)
            else:
                raise JSONQueryException(f"Bad query: {action}")


class Q:
    """
    A Query object. We can join different queries by means of bitand and bitor operators (& |)
    Examples
    --------
    >> obj = JSONObject(
        [
            {
                "timestamp": "2021-05-01 09:00:00",
                "value": [0.5, 0.87]
            },
            {
                "timestamp": "2021-04-02 10:30:00",
                "value": [-0.23, 1]
            },
            {
                "timestamp": "2021-06-01 08:25:30",
                "value": [0.9, 0.15]
            }
        ]
    )

    >> obj.query(Q(timestamp__gt="2021-05-01 10:00:00") | Q(value__0__gte=0.5))
        [{"timestamp": "2021-05-01 09:00:00","value": [0.5, 0.87]},{"timestamp": 2021-06-01 08:25:30, "value": [0.9, 0.15]}]
    """

    # TODO make Q object
    def __init__(self, **kwargs):

        self.AND = []
        self.OR = []
        self.NOT = []

        if len(kwargs) > 0:
            self._parse_args(kwargs)

    def _parse_args(self, kwargs):
        for k, v in kwargs.items():

            splitted = *k.split("__"), v
            self.AND.append(splitted)

    def __and__(self, other):

        if not isinstance(other, Q):
            raise TypeError(f"Cannot add instances of {type(self)} and {type(other)}")

        obj = Q()
        obj.AND = self.AND + other.AND
        return obj

    def __or__(self, other):

        if not isinstance(other, Q):
            raise TypeError(f"Cannot add instances of {type(self)} and {type(other)}")

        obj = Q()
        obj.OR = self.OR + other.OR
        return obj


class QuerySet(list):
    """
    This is a queryset object.
    Attributes:
    ----------
        _root: the root json object from which the entire queryset is derived
    """

    def __init__(self, *args):
        super().__init__(*args)
        self._root = None

    def first(self):
        return self.__getitem__(0) if self.__len__() > 0 else None

    def last(self):
        return self.__getitem__(-1) if self.__len__() > 0 else None

    def exists(self):
        return True if self.__len__() > 0 else False

    def count(self):
        return self.__len__()

    def update(self, new_obj):
        """
        Update elements of queryset within JSONObject from which they are derived (self._root)
        """

        # TODO add test for update when callables
        is_callable = callable(new_obj)

        for item in self:
            path = item.jsonpath.relative_to(self._root)
            if is_callable:
                try:
                    exec(f"self._root{path} = new_obj(self._root{path})")
                except Exception:
                    pass
            else:
                exec(f"self._root{path} = new_obj")

    def unique(self):
        """
        Returns unique values in a querylist
        """
        from jsonutils.base import JSONSingleton

        # TODO add test for this
        # TODO dict option for counting unique values
        unique_values = QuerySet()
        unique_values._root = self._root
        for idx, item in enumerate(self):
            if item not in unique_values:
                unique_values.append(item)
        return unique_values

    def filter(self, **q):
        # TODO add test
        output = QuerySet()
        output._root = self._root
        for item in self:
            if item.query(**q).exists():
                output.append(item)
        return output

    def __repr__(self):
        return "<QuerySet " + super().__repr__() + ">"


class AllChoices(type):
    pass


class All(metaclass=AllChoices):
    """
    When we request this object as a target query value, we retrieve all objects of given key

    Examples:
    --------

    >> test = JSONObject(
        {
            "A": [
                {
                    "A": 1,
                    "B": 2
                },
                {
                    "A": 2,
                    "B": 3
                }
            ]
        }
    )

    >> test.query(A=All)
        [[{"A": 1, "B": 2},{"A": 2, "B": 3}], 1, 2]
    """

    pass



"""This module provides json serializers for the json objects"""

from json import JSONEncoder


class JSONObjectEncoder(JSONEncoder):
    """
    We need this custom encoder in order to be able to use json.dumps on a JSONObject instance, due to the presence of JSONBool type,
    which is not JSON serializable
    """

    def default(self, o):
        from jsonutils.base import JSONBool, JSONNull

        if isinstance(o, (JSONBool, JSONNull)):
            return o._data
        return super().default(o)



import random
import string

from jsonutils.base import (
    JSONBool,
    JSONDict,
    JSONFloat,
    JSONInt,
    JSONList,
    JSONNull,
    JSONStr,
)


def dummy_json(
    min_dict_length=2,
    max_dict_length=5,
    min_list_length=2,
    max_list_length=5,
    min_str_length=10,
    max_str_length=100,
    max_depth=10,
):
    """
    Generate a dummy JSONObject instance
    """
    # TODO

    # assert argument values are correct
    assert (
        min_dict_length <= max_dict_length
    ), "Argument min_dict_length must be less or equal than max_dict_length"

    assert (
        min_list_length <= max_list_length
    ), "Argument min_list_length must be less or equal than max_list_length"

    assert (
        min_str_length <= max_str_length
    ), "Argument min_str_length must be less or equal than max_str_length"

    NODE_LIST_ALL = (
        JSONBool,
        JSONDict,
        JSONFloat,
        JSONInt,
        JSONList,
        JSONNull,
        JSONStr,
    )

    NODE_LIST_COMPOSED = (
        JSONDict,
        JSONList,
    )

    NODE_LIST_SINGLETON = (JSONStr, JSONFloat, JSONInt, JSONBool, JSONNull)

    def get_random_string(N=4):
        return "".join(random.choices(string.ascii_lowercase, k=N))

    def get_random_node(kind="all"):
        if kind == "all":
            return random.choice(NODE_LIST_ALL)
        if kind == "composed":
            return random.choice(NODE_LIST_COMPOSED)
        if kind == "singleton":
            return random.choice(NODE_LIST_SINGLETON)
        else:
            raise ValueError(
                "Argument kind can only be some of the following: 'all', 'composed' or 'singletion'"
            )

    def initialize_singleton(node):
        "node must be a JSONSingleton instance"

        if node == JSONStr:
            out = get_random_string(8)
            return JSONStr(out)
        elif node == JSONBool:
            out = random.choice((True, False))
            return JSONBool(out)
        elif node == JSONInt:
            out = random.randint(0, 10000)
            return JSONInt(1)
        elif node == JSONFloat:
            out = random.random()*100
            return JSONFloat(out)
        elif node == JSONNull:
            return JSONNull(None)
        else:
            raise TypeError(f"node argument is not a JSONSingleton instance: {node}")

    def fill_composed_node(node, it=0):
        "node must be a JSONCompose instance"

        if isinstance(node, dict):
            length = random.randint(min_dict_length, max_dict_length)
            for _ in range(length):
                key = get_random_string()
                if it < max_depth:
                    value = get_random_node(kind="all")  # a node object
                else:
                    value = get_random_node(kind="singleton")
                if value.is_composed:
                    value = value()
                    fill_composed_node(value, it + 1)
                else:
                    value = initialize_singleton(value)
                node.__setitem__(key, value)
        elif isinstance(node, list):
            length = random.randint(min_list_length, max_list_length)
            for _ in range(length):
                if it < max_depth:
                    item = get_random_node(kind="all")
                else:
                    item = get_random_node(kind="singleton")
                if item.is_composed:
                    item = item()
                    fill_composed_node(item, it + 1)
                else:
                    item = initialize_singleton(item)
                node.append(item)

    # get the initial node
    initial_node = get_random_node(kind="composed")()

    fill_composed_node(initial_node)
    return initial_node



"""
We define here the interactions between a query and a particular node.
Nodes can be of the following types:
    JSONList, JSONDict, JSONStr, JSONBool, JSONInt, JSONFloat, JSONNull
whereas query target values can be:
    list/tuple, dict, str, date/datetime, bool, int/float, null

So we have up to 49 checks for each action:
============================
  NODE     |    QUERY_VALUE
============================
JSONList -----> list/tuple
         -----> dict
         -----> str
         -----> date/datetime
         -----> bool
         -----> float/int 
         -----> null

JSONDict -----> list/tuple
         -----> dict
         -----> str
         -----> date/datetime
         -----> bool
         -----> float/int
         -----> null
    .
    .
    .
"""

import re
from datetime import date, datetime

import pytz
from jsonutils.base import (
    JSONBool,
    JSONCompose,
    JSONDict,
    JSONFloat,
    JSONInt,
    JSONList,
    JSONNull,
    JSONSingleton,
    JSONStr,
)
from jsonutils.exceptions import JSONQueryException
from jsonutils.functions.parsers import parse_datetime, parse_float
from jsonutils.query import All


def _gt(node, requested_value):
    """
    Greather than action

           \ req_value  |  dict  |  list/tuple  |  bool  |  float/int  |  str  |  datetime  |  null
      node  \           |        |              |        |             |       |            |
    ================================================================================================
           JSONDict     |    X   |       X      |    X   |      X      |   X   |     X      |   X
                        |        |              |        |             |       |            |
           JSONList     |    X   |       V      |    X   |      X      |   X   |     X      |   X
                        |        |              |        |             |       |            |
           JSONStr      |    X   |       X      |    X   |      V      |   V   |     V      |   X
                        |        |              |        |             |       |            |
           JSONBool     |    X   |       X      |    X   |      X      |   X   |     X      |   X
                        |        |              |        |             |       |            |
           JSONInt      |    X   |       X      |    X   |      V      |   V   |     X      |   X
                        |        |              |        |             |       |            |
           JSONFloat    |    X   |       X      |    X   |      V      |   V   |     X      |   X
                        |        |              |        |             |       |            |
           JSONNull     |    X   |       X      |    X   |      X      |   X   |     X      |   X
    """

    if isinstance(node, JSONList):
        if isinstance(requested_value, (list, tuple)):
            requested_value = list(requested_value)

            len_node = len(node)
            len_requ = len(requested_value)
            length = min((len_node, len_requ))
            success_number = 0
            control_success = False
            for i in range(length):
                if node[i] > requested_value[i]:
                    success_number += 1
                    control_success = True
            if success_number == length and control_success:
                return True
            else:
                return False
    elif isinstance(node, JSONDict):
        return False
    elif isinstance(node, JSONSingleton):
        # if dealing with singletons, call its rich comparison methods
        return node > requested_value


def _gte(node, requested_value):
    """
    Greather-equal than action. Matches are the same as _gt action
    """
    if isinstance(node, JSONList):
        if isinstance(requested_value, (list, tuple)):
            requested_value = list(requested_value)

            len_node = len(node)
            len_requ = len(requested_value)
            length = min((len_node, len_requ))
            success_number = 0
            control_success = False
            for i in range(length):
                if node[i] >= requested_value[i]:
                    success_number += 1
                    control_success = True
            if success_number == length and control_success:
                return True
            else:
                return False
    elif isinstance(node, JSONDict):
        return False
    elif isinstance(node, JSONSingleton):
        # if dealing with singletons, call its rich comparison methods
        return node >= requested_value


def _lt(node, requested_value):
    """
    Lower than action. Matches are the same as _gt action
    """
    if isinstance(node, JSONList):
        if isinstance(requested_value, (list, tuple)):
            requested_value = list(requested_value)

            len_node = len(node)
            len_requ = len(requested_value)
            length = min((len_node, len_requ))
            success_number = 0
            control_success = False
            for i in range(length):
                if node[i] < requested_value[i]:
                    success_number += 1
                    control_success = True
            if success_number == length and control_success:
                return True
            else:
                return False
    elif isinstance(node, JSONDict):
        return False
    elif isinstance(node, JSONSingleton):
        # if dealing with singletons, call its rich comparison methods
        return node < requested_value


def _lte(node, requested_value):
    """
    Lower-equal than action. Matches are the same as _gt action
    """
    if isinstance(node, JSONList):
        if isinstance(requested_value, (list, tuple)):
            requested_value = list(requested_value)

            len_node = len(node)
            len_requ = len(requested_value)
            length = min((len_node, len_requ))
            success_number = 0
            control_success = False
            for i in range(length):
                if node[i] <= requested_value[i]:
                    success_number += 1
                    control_success = True
            if success_number == length and control_success:
                return True
            else:
                return False
    elif isinstance(node, JSONDict):
        return False
    elif isinstance(node, JSONSingleton):
        # if dealing with singletons, call its rich comparison methods
        return node <= requested_value


def _exact(node, requested_value):
    """
    An exact match.
    In an exact match, only same types are checked, except for JSONStr, which is more versatile.

           \ req_value  |  dict  |  list/tuple  |  bool  |  float/int  |  str  |  datetime  |  null
      node  \           |        |              |        |             |       |            |
    ================================================================================================
           JSONDict     |    V   |       X      |    X   |      X      |   X   |     X      |   X
                        |        |              |        |             |       |            |
           JSONList     |    X   |       V      |    X   |      X      |   X   |     X      |   X
                        |        |              |        |             |       |            |
           JSONStr      |    X   |       X      |    V   |      V      |   V   |     V      |   X
                        |        |              |        |             |       |            |
           JSONBool     |    X   |       X      |    V   |      X      |   V   |     X      |   X
                        |        |              |        |             |       |            |
           JSONInt      |    X   |       X      |    X   |      V      |   V   |     X      |   X
                        |        |              |        |             |       |            |
           JSONFloat    |    X   |       X      |    X   |      V      |   V   |     X      |   X
                        |        |              |        |             |       |            |
           JSONNull     |    X   |       X      |    X   |      X      |   X   |     X      |   V
    """

    if requested_value == All:
        return True

    if isinstance(node, JSONDict):
        if isinstance(requested_value, dict):
            return node == requested_value
        else:
            return False
    elif isinstance(node, JSONList):
        if isinstance(requested_value, (list, tuple)):
            return node == list(requested_value)
        else:
            return False
    elif isinstance(node, JSONSingleton):
        return node == requested_value
    else:
        return False


def _contains(node, requested_value):
    """
    This method analyzes whether a given JSONObject contains the object specified by the <other> parameter, and returns a boolean.
    <self> will be the current child instance within the JSONObject, whereas <other> will be the current query target value.
    Examples in a query method:
    --------------------------
    >> obj = JSONObject(
        {
            "data": {
                "cik": "0008547852",
                "country": "USA",
                "live": False
            },
            "team": [
                "Daniel",
                "Alex",
                "Catherine",
                None
            ]
        }
    )

    >> obj.query(data__contains="cik")  # in this case the "cik" string object will play the role of <other> (the target query value).
                                        # On the other hand, <self> in this case will take the value of the "data" dictionary (JSONDict)
        [{'cik': '0008547852', 'country': 'USA', 'live': False}]

    >> obj.query(cik__contains=85) # now the 85 integer object will be <other> object, whereas <self> will be the "cik" string object.
        ['0008547852']

    >> obj.query(team__contains=["Alex", "Daniel"]) # in this case target JSONObject must contains both "Alex" and "Daniel" strings
        [['Daniel', 'Alex', 'Catherine', None]]

    >> obj.query(team__contains=None).first()
        ['Daniel', 'Alex', 'Catherine', None]
    """

    # TODO implement clever parsing
    if isinstance(node, JSONStr):
        # if target object is an string, contains will return True if target value/s are present within it.
        if isinstance(requested_value, str):
            return requested_value in node
        elif isinstance(requested_value, (float, int)):
            # if target value is a number, we convert it first to a string and then check if it is present within self
            return str(requested_value) in node
        elif isinstance(requested_value, (list, tuple)):
            # if target value is a list, then check if all its items are present in self string
            return all(str(x) in node for x in requested_value)
    elif isinstance(node, JSONDict):
        # if target object is a dict, contains will return True if target value/s are present within its keys.
        if isinstance(requested_value, str):
            return requested_value in node.keys()
        elif isinstance(requested_value, (list, tuple)):
            return all(x in node.keys() for x in requested_value)
    elif isinstance(node, JSONList):
        # if target object is a list, contains will return True if target value are present within its elements.
        if isinstance(requested_value, (list, tuple)):
            return all(x in node for x in requested_value)
        else:
            return True if requested_value in node else False
    elif isinstance(node, JSONBool):
        if isinstance(requested_value, bool):
            return node._data == requested_value
    elif isinstance(node, JSONNull):
        if isinstance(requested_value, type(None)):
            return node._data == requested_value
    elif isinstance(node, (JSONFloat, JSONInt)):
        if isinstance(requested_value, (int, float, str)):
            return str(requested_value) in str(node)
    return False


def _icontains(node, requested_value):
    """
    Case insensitive version of contains
    """

    # TODO implement clever parsing
    # TODO if self is a number and other too
    if isinstance(node, JSONStr):
        # if target object is an string, contains will return True if target value/s are present within it.
        if isinstance(requested_value, str):
            return requested_value.lower() in node.lower()
        elif isinstance(requested_value, (float, int)):
            # if target value is a number, we convert it first to a string and then check if it is present within self
            return str(requested_value) in node
        elif isinstance(requested_value, (list, tuple)):
            # if target value is a list, then check if all its items are present in self string
            try:
                return all(str(x).lower() in node.lower() for x in requested_value)
            except Exception:  # maybe if x has a unknown type # TODO possible conflict if x is an object and its str representation matches node
                return False
    elif isinstance(node, JSONDict):
        # if target object is a dict, contains will return True if target value/s are present within its keys.
        if isinstance(requested_value, str):
            try:
                return requested_value.lower() in map(str.lower, node.keys())
            except Exception:  # maybe JSONDict has integer keys. TODO handle this
                return False
        elif isinstance(requested_value, (list, tuple)):
            try:
                return all(
                    str(x).lower() in map(str.lower, node.keys())
                    for x in requested_value
                )
            except Exception:
                return False
    elif isinstance(node, JSONList):
        # if target object is a list, contains will return True if target value are present within its elements.
        # TODO modify this, must compare lower strings and ignore other objects
        if isinstance(requested_value, (list, tuple)):
            return all(x in node for x in requested_value)
        else:
            return True if requested_value in node else False
    elif isinstance(node, JSONBool):
        if isinstance(requested_value, bool):
            return node._data == requested_value
    elif isinstance(node, JSONNull):
        if isinstance(requested_value, type(None)):
            return node._data == requested_value
    return False


def _in(node, requested_value):
    # TODO complete in child method
    """
    This method, as opposed to "contains", analyzes whether a given JSONObject is contained in the iterable object specified
    by the <other> parameter
    """

    if isinstance(node, (JSONSingleton)):
        # <self> might be JSONStr, JSONFloat, JSONInt, JSONBool or JSONNull.
        if isinstance(requested_value, (str, list, tuple, dict)):
            return node in requested_value
    elif isinstance(node, JSONList):
        if isinstance(requested_value, (list, tuple)):
            return all(x in requested_value for x in node)
    elif isinstance(node, JSONDict):
        if isinstance(requested_value, (list, tuple)):
            return all(x in requested_value for x in node.keys())
    return False


def _regex(node, requested_value):
    """
    This method analyzes whether a given JSONObject matchs with target regex pattern specified by <other>.
    """

    if isinstance(node, JSONStr):
        if isinstance(requested_value, (str, re.Pattern)):
            return bool(re.search(requested_value, node))
    elif isinstance(node, (JSONInt, JSONFloat)):
        if isinstance(requested_value, (str, re.Pattern)):
            return bool(re.search(requested_value, str(node)))
    return False


def _fullregex(node, requested_value):
    """
    This method analyzes whether a given JSONObject full matchs with target regex pattern specified by <other>.
    """
    if isinstance(node, JSONStr):
        if isinstance(requested_value, (str, re.Pattern)):
            return bool(re.fullmatch(requested_value, node))
    elif isinstance(node, (JSONInt, JSONFloat)):
        if isinstance(requested_value, (str, re.Pattern)):
            return bool(re.fullmatch(requested_value, str(node)))
    return False


def _isnull(node, requested_value):
    """
    This method analyzes whether a given JSONObject is null.
    """
    if not isinstance(requested_value, bool):
        raise JSONQueryException(
            f"Requested value must be a boolean, not {type(requested_value)}"
        )
    return bool(node) != requested_value if node != 0 else False


def _length(node, requested_value):
    """
    This method analyzes whether a given JSONObject has requested length.
    """
    if not isinstance(requested_value, int):
        raise JSONQueryException(
            f"Requested value must be an int, not {type(requested_value)}"
        )
    if isinstance(node, (JSONCompose, JSONStr)):
        return node.__len__() == requested_value
    elif isinstance(node, JSONInt):
        return str(node).__len__() == requested_value
    else:
        return False


def _type(node, requested_value):
    """
    This method analyzes whether a given JSONObject has requested type.
    """
    if requested_value not in (
        dict,
        list,
        str,
        datetime,
        float,
        int,
        bool,
        None,
        "dict",
        "list",
        "str",
        "datetime",
        "float",
        "int",
        "bool",
        "None",
    ):
        raise JSONQueryException(
            f"Requested value must be a valid type, not {requested_value}"
        )

    if isinstance(node, JSONBool):
        if requested_value in (bool, "bool"):
            return True
        else:
            return False
    elif isinstance(node, JSONNull):
        if requested_value in (None, "None"):
            return True
        else:
            return False
    else:
        return isinstance(node, requested_value)


def _key(node, requested_value):
    """
    This method analyzes whether a given JSONObject comes from a given parent key
    """
    # TODO add test
    if not isinstance(requested_value, str):
        raise TypeError(
            f"Argument requested_value must be an str instance, not {type(requested_value)}"
        )
    return node._key == requested_value


def _index(node, requested_value):
    """
    This method analyzes whether a given JSONObject comes from a given parent key
    """
    # TODO add test
    if not isinstance(requested_value, int):
        raise TypeError(
            f"Argument requested_value must be an int instance, not {type(requested_value)}"
        )
    return node._index == requested_value


def _path(node, requested_value):
    """
    This method analyzes whether a given JSONObject comes from selected path
    """
    # TODO store jsonpath in cache to avoid call it repeatdly
    # TODO add test
    if isinstance(requested_value, (list, tuple, set)):
        return all(i in node.jsonpath.keys for i in requested_value)
    elif isinstance(requested_value, (str, int)):
        return requested_value in node.jsonpath.keys
    else:
        raise TypeError(
            f"Argument requested_value must be an iterable, str or bool, not {type(requested_value)}"
        )



class JSONQueryException(Exception):
    """Base class for JSON Query exceptions"""

    pass


class JSONSingletonException(Exception):
    pass


class JSONDecodeException(Exception):
    pass

class JSONQueryMultipleValues(JSONQueryException):
    pass



import requests
import time

class default:
    pass


def retry_function(
    func, *args, max_tries=10, output_value=default, raise_exception=True, **kwargs
):
    """
    Try to evaluate selected function a certain number of tries
    """
    out = None
    e = None
    for _ in range(max_tries):
        try:
            out = func(*args, **kwargs)
        except Exception as exc:
            e = exc
            continue
        else:
            if func in (
                requests.post,
                requests.get,
                requests.put,
                requests.patch,
                requests.delete,
            ):
                if hasattr(out, "status_code"):
                    if out.status_code > 299:
                        time.sleep(0.5)
                        continue
                    else:
                        return out if output_value == default else output_value
            else:
                return out if output_value == default else output_value

    if raise_exception is True:
        try:
            msg = e or out.text
        except Exception:
            msg = e or out
        raise Exception(
            f"Selected function does not give any valid output after {max_tries} tries.\nError message: {msg}\n"
        )
    else:
        return out if output_value == default else output_value